-- everything for espio in battle mod

-- command to switch how invis controls
COM_AddCommand(espio.InvisModeCommand, function(p, mode)
	p.espio_invismode = $ or 1	-- make sure this exists
	
	if not (mode)
		local modes = {"Hold", "Toggle"}
		CONS_Printf(p, "\"espioinvismode\" is \""..modes[p.espio_invismode].."\" default is \"Hold\"")
		return
	else
		if (tonumber(mode))	-- number handling
			mode = tonumber($)
			if (mode != 1 and mode != 2)	-- only 2 possible values so I can use this
				CONS_Printf(p, "\""..mode.."\" is not a possible value for \"espioinvismode\"")
				return
			end
			
			p.espio_invismode = mode
			espio.WriteConfig(p)	-- update config
		else	-- string handling
			mode = mode:lower()
			if (mode != "hold" and mode != "toggle")
				CONS_Printf(p, "\""..mode.."\" is not a possible value for \"espioinvismode\"")
				return
			end
			
			p.espio_invismode = (mode == "hold") and 1 or 2
			espio.WriteConfig(p)	-- update config
		end
	end
end)

-- loading espio's stuff
local loaded = false
espio.BattleSetup = function()
	if not (CBW_Battle) then return end	-- battle isnt loaded
	if (loaded) then return end	-- already setup for battle? dont run
	local B = CBW_Battle
	local S = B.SkinVars
	
	-- espio's priority
	local function espio_battlePriority(p)
		if (p.actionstate == 2)
			-- espio should have no attacking power himself, the tornado does the damaging here
			local atk, def = 0, 2
			B.SetPriority(p, atk, def, nil, atk, def, "leaf swirl")
		else
			local spinjump = ((p.pflags & PF_JUMPED) and not (p.pflags & PF_NOJUMPDAMAGE))
			local spinning = (p.pflags & PF_SPINNING)
			
			if (spinjump or spinning)
				-- make espio have a drill spin instead of a spin attack
				local atk, def = 1, 1
				local text = (spinning) and "drill-spin attack" or "jumping drill-spin attack"
				B.SetPriority(p, atk, def, nil, atk, def, text)
			end
		end
	end
	
	-- separate function for leaf swirl functionality, just so I can return on the end conditions
	local function espio_leafSwirl(p)
		local mo = p.mo
		
		-- end conditions
		if (p.actiontime == espio.LeafSwirlDuration)	-- duration is up
		or (P_PlayerInPain(p))	-- we're in pain
			if not (P_PlayerInPain(p))	-- we wouldnt want to set any state here
				mo.state = (P_IsObjectOnGround(mo)) and S_PLAY_STND or S_PLAY_FALL
			end
			
			p.actionstate = 0
			mo.espio_leafswirl_parryhack = 3	-- need this here too
			return
		end
				
		if (mo.state != S_ESPIO_LEAFSWIRL) then mo.state = S_ESPIO_LEAFSWIRL end
		local frametics = espio.LeafSwirlFrameTics
				
		-- get frame
		local frame
		local tics = 0
		for i = 0, #frametics-1
			tics = $+frametics[i+1]
			if (p.actiontime <= tics)
				-- only reduce momentum before the final frame
				local moms = {"momx", "momy", "momz"}
				for i = 1, 3
					local mom = moms[i]
					mo[mom] = FixedMul($, espio.LeafSwirlMultiplier)
				end
						
				frame = i
				break
			end
		end
		if (frame == nil) then frame = 6 end	-- assume we're on the final frame
		espio.SetFrame(mo, frame)
				
		-- final frame is attack time
		if (frame == 6)
			mo.espio_leafswirl_parryhack = 3	-- we'll need this to get parried!
			
			-- spawn the tornado
			if (p.actiontime == (tics + 1))
				local tornado = P_SpawnMobjFromMobj(mo, 0, 0, FixedMul(mo.scale, espio.LeafSwirlTornadoSpawnOffset), MT_ESPIO_LEAFSWIRL_TORNADO)
				tornado.target = mo	-- we'll want to save this to know who spawned it
				
				-- make sure it's at least always on the floor
				local checkz = tornado.z
				if (mo.eflags & MFE_VERTICALFLIP)
					checkz = tornado.z + tornado.height
				end
				
				if ((mo.eflags & MFE_VERTICALFLIP) and checkz > P_CeilingzAtPos(mo.x, mo.y, mo.z, mo.height))
				or (not (mo.eflags & MFE_VERTICALFLIP) and checkz < P_FloorzAtPos(mo.x, mo.y, mo.z, mo.height))
					local newz = P_FloorzAtPos(mo.x, mo.y, mo.z, mo.height)
					if (mo.eflags & MFE_VERTICALFLIP)
						newz = P_CeilingzAtPos(mo.x, mo.y, mo.z, mo.height)
					end
					
					P_TeleportMove(tornado, tornado.x, tornado.y, newz)
				end
			end
			
			-- remove all momentum at this point
			mo.momx, mo.momy, mo.momz = 0, 0, 0
		end
	end
	local function espio_battleSpecial(mo, doaction)
		local p = mo.player
		local grounded = P_IsObjectOnGround(mo)
		
		-- ability names
		if (mo.espio_chargejump)
			p.actiontext = "Charge Thrust"
			p.actionrings = 10
		else
			p.actiontext = (grounded or (p.actionstate == 1)) and "Invisibility" or "Leaf Swirl"
			p.actionrings = (grounded or (p.actionstate == 1)) and 5 or 10
		end
		p.action2text = nil	-- this'll get changed later if we're invisible
		
		if (p.actionstate == 0)	-- starting the move?
			if (doaction == 1)	-- just pressed it
				B.PayRings(p)
				
				if (mo.espio_chargejump)	-- charge thrust
					local timer = mo.espio_chargejump_timer
					
					-- calculate speed, then launch!
					local range = (espio.ChargeThrustMax - espio.ChargeThrustMin)	-- we'll need the range for calcs
					local charge = min(espio.ChargeThrustMax, espio.ChargeThrustMin + ((range/espio.ChargeJumpDuration)*timer))
					
					local thrust = FixedMul(mo.scale, charge/((mo.eflags & MFE_UNDERWATER) and 2 or 1))
					mo.state = S_PLAY_ROLL
					mo.momz = 0	-- cancel z momentum for a full horizontal thrust
					P_InstaThrust(mo, mo.angle, thrust)
					S_StartSound(mo, sfx_kc5b)
					//print("Thrust: "..charge/FRACUNIT)
					
					-- apply cooldown
					local range = (espio.ChargeThrustMaxCooldown - espio.ChargeThrustMinCooldown)
					local cooldown = min(espio.ChargeThrustMaxCooldown, espio.ChargeThrustMinCooldown + ((timer*espio.ChargeJumpDuration)/range))
					B.ApplyCooldown(p, cooldown)
					
					-- visuals
					local dusts = 18
					for i = 1, dusts
						local radius, height = mo.radius>>FRACBITS, mo.height>>FRACBITS
						local x = P_RandomRange(-(radius), radius)*FRACUNIT
						local y = P_RandomRange(-(radius), radius)*FRACUNIT
						local z = P_RandomRange(0, height)*FRACUNIT
						
						local dust = P_SpawnMobjFromMobj(mo, x, y, z, MT_SPINDUST)
						if (mo.eflags & MFE_UNDERWATER) 
							dust.state = S_WPLD1 
							dust.scale = $/2
						end
					end
					mo.espio_chargejump_visualtics = espio.ChargeJumpVisualTics
					
					-- reset vars
					espio.StopChargeSounds(mo)
					espio.ResetChargeJump(mo)
				else
					p.actionstate = (grounded) and 1 or 2	-- invisibility is 1, leaf swirl is 2
					p.actiontime = 0	-- timer for abilities
					p.canguard = false
					
					-- init functions on each state
					if (p.actionstate == 1)	-- invisibility
						p.normalspeed = FixedMul(espio.InvisWalkMultiplier, skins[mo.skin].normalspeed)
						mo.espio_ringdraintics = espio.InvisRingDrainTicBounds[2]
						if (p.rings) then mo.flags2 = $|(MF2_FRET) end	-- prevent targeting
						
						-- set opacity!
						mo.espio_battleopacity = (p.rings == 0) and (espio.InvisNoRingFade * 10) or 0
						
						-- visuals
						local dusts = 42
						for i = 1, dusts
							local radius, height = (mo.radius*9/4)>>FRACBITS, (mo.height)>>FRACBITS
							local x = P_RandomRange(-(radius), radius)*FRACUNIT
							local y = P_RandomRange(-(radius), radius)*FRACUNIT
							local z = P_RandomRange(0, height)*FRACUNIT
							
							local dust = P_SpawnMobjFromMobj(mo, x, y, z, MT_SPINDUST)
							local savescale = dust.scale
							if (mo.eflags & MFE_UNDERWATER)
								dust.state = S_WPLD1 
								dust.scale = $/2
							else
								dust.tics = $/2
							end
							dust.momz = FixedMul(savescale, P_RandomRange(2, 5)*FRACUNIT) * P_MobjFlip(mo)
						end
						espio.SpawnSoundMobj(mo, sfx_s3k82)	-- just so the sound doesnt follow us!
					else	-- leaf swirl
						p.pflags = $ & ~(PF_JUMPED|PF_SHIELDABILITY)
						p.secondjump = 0
						p.climbing = 0	-- get off the wall!
						mo.state = S_ESPIO_LEAFSWIRL
						B.ApplyCooldown(p, espio.LeafSwirlCooldown)
					end
				end
			end
		else	-- ability thinker
			p.actiontime = $+1
			p.canguard = false
			
			if (p.actionstate == 1)	-- invisibility
				-- walk slower while invisible
				p.normalspeed = FixedMul(espio.InvisWalkMultiplier, skins[mo.skin].normalspeed)
				mo.espio_fixnormalspeed = true	-- we'll need this
				
				-- instantly go to waiting while invisible
				if (mo.state == S_PLAY_STND) then mo.state = S_PLAY_WAIT end
				
				-- drain rings while we're invisible based on velocity
				local speedthreshold = FixedMul(p.normalspeed, espio.InvisWalkThreshold)/FRACUNIT
				local speed = FixedHypot(mo.momx, mo.momy)/FRACUNIT
				
				local ticbounds = espio.InvisRingDrainTicBounds
				local ticrange = ticbounds[2] - ticbounds[1]
				local mintics = max(1, ticbounds[2] - min(ticrange, (speed * ticrange) / speedthreshold))
				
				mo.espio_ringdraintics = min(mintics, $) - 1
				if not (mo.espio_ringdraintics)
					mo.espio_ringdraintics = mintics	-- reset timer
					P_GivePlayerRings(p, -1)
				end
				
				-- end conditions
				p.espio_invismode = $ or 1	-- make sure this exists
				
				if ((p.espio_invismode == 1) and not doaction)	-- HOLD to keep invis on
				or ((p.espio_invismode == 2) and (doaction == 1))	-- TOGGLE to disable invis
					p.actionstate = 0
					mo.flags2 = $ & ~(MF2_FRET)	-- let us be targeted again
					B.ResetPlayerProperties(p)
					B.ApplyCooldown(p, espio.InvisCooldown)
					S_StartSound(mo, sfx_prloop)
				end
			else	-- leaf swirl
				espio_leafSwirl(p)	-- easier to just have it in a function
			end
		end
		
		-- handle the leaf swirl hack
		if (mo.espio_leafswirl_parryhack)
			mo.espio_leafswirl_parryhack = $-1
		end
		
		-- get how opaque we should be based on momentum
		if (p.actionstate == 1)	-- we're in the invisibility state, adjust based on momentum!
			-- speed values
			local speedthreshold = FixedMul(p.normalspeed, espio.InvisWalkThreshold)/FRACUNIT
			local speed = FixedHypot(mo.momx, mo.momy)/FRACUNIT
			
			-- set how close we are to being visible as our action2text
			if (p.rings == 0)	-- we're visible when we have no rings regardless
				p.action2text = "Visible!"
				p.action2textflags = (leveltime&4 <= 2) and 2 or 0	-- TF_YELLOW = 2 (these arent exposed)
				mo.flags2 = $ & ~(MF2_FRET)	-- let us be targeted again
			else
				local speedvalue = min(100, speed*100/speedthreshold)
				p.action2text = "Speed "..speedvalue.."%"
				p.action2textflags = (speedvalue == 100 and (leveltime&4 <= 2)) and 2 or 0	-- TF_YELLOW = 2 (these arent exposed)
				mo.flags2 = $|(MF2_FRET)	-- prevent targeting
			end
			
			-- now get the target opacity
			local targetopacity = ((p.rings == 0) and espio.InvisNoRingFade or 0) * 10
			if (mo.espio_battleopacity < targetopacity)
				mo.espio_battleopacity = min(targetopacity, $ + espio.InvisTransparencyAdjust)
			elseif (mo.espio_battleopacity > targetopacity)
				mo.espio_battleopacity = max(targetopacity, $ - espio.InvisTransparencyAdjust)
			end
		else
			-- might as well fix normalspeed here
			if (mo.espio_fixnormalspeed)
				B.ResetPlayerProperties(p)
				mo.espio_fixnormalspeed = nil
			end
			
			-- bring espio's transparency back to normal
			if (mo.espio_battleopacity != nil)
				mo.espio_battleopacity = min(100, $ + espio.InvisTransparencyAdjust)
				if (mo.espio_battleopacity == 100)
					mo.espio_battleopacity = nil
					
					-- reset this while we're at it
					mo.frame = ($ & ~FF_TRANSMASK)
					mo.flags2 = $ & ~(MF2_DONTDRAW)
				end
			end
		end
		
		-- set espio's transparency
		if (mo.espio_battleopacity != nil)
			-- now set our transparency!
			local trans = 10 - (mo.espio_battleopacity/10)
			//print("Transparency: "..trans)
			if (trans == 10)	-- fully transparent, use MF2_DONTDRAW
				mo.frame = ($ & ~FF_TRANSMASK) 	-- remove transparency, we wont need it here
				mo.flags2 = $|(MF2_DONTDRAW)
			else
				mo.frame = ($ & ~FF_TRANSMASK)
				mo.flags2 = $ & ~(MF2_DONTDRAW)
				if (trans >= 1 and trans <= 9)
					mo.frame = $|(trans << FF_TRANSSHIFT)
				end
			end
		end
	end
	
	-- espio exhaust
	local espio_battleExhaust = function(p)
		if not (espio.AbilityCheck(p)) then return end
		local mo = p.mo
			
		-- conditions to lower it
		if (mo.espio_ceilingCling)
		or (mo.espio_wallrun)
		or (mo.espio_chargejump)
			local maxtime = espio.BattleExhaustTime
			p.exhaustmeter = max(0, $- (FRACUNIT/maxtime))
		end
		
		-- refill here
		if ((P_IsObjectOnGround(mo) and not mo.espio_ceilingCling) or P_PlayerInPain(p))
			p.exhaustmeter = FRACUNIT
		end
	
		return true
	end
	
	-- setup espio's skinvars
	S["espio"] = {
		weight = 90,
		special = espio_battleSpecial,
		func_priority_ext = espio_battlePriority,
		func_exhaust = espio_battleExhaust,	-- CE based
		guard_frame = 6,
	}
	
	-- slight edit to B.DummyLockOn to fix homing invisible espio players!
	B.DummyLockOn = function(mo,target)
		mo.height = target.height
		mo.scale = target.scale
		mo.flags = $|MF_NOGRAVITY
		if target.player.playerstate == PST_LIVE then
			B.DummyTeleport(mo) //Reposition
		end
		
		//Get aimable
		if target.player.playerstate == PST_LIVE and not target.player.intangible
		and not (target.flags & MF_NOCLIPTHING)
		and not (target.flags2 & MF2_FRET)	-- espio fix
			mo.flags2 = $|MF2_INVERTAIMABLE //Allow lock-on
		else
			//Target is not alive. Disallow lock-on.
			mo.flags2 = $&~MF2_INVERTAIMABLE
		end
	end
	
	loaded = true
end

-- hide overlays while we're invisible
-- (gotta do this outside the function since you cant add hooks within loops/coroutines)
local function espio_hideOverlay(overlay)
	local owner = overlay.target
	if not (owner and owner.valid) then return end
	if not (owner.player) then return end	-- this is needed for MT_OVERLAY apparently
		
	if (owner.espio_battleopacity != nil) 
		overlay.flags2 = $|(MF2_DONTDRAW)
		if (overlay.tracer and overlay.tracer.valid)
			-- the overlay has a 2nd layer (like the s3k shields!)
			overlay.tracer.flags2 = $|(MF2_DONTDRAW)
		end
		
		return true 
	end
	
	-- if we're here, we have no need for it to not be drawn
	overlay.flags2 = $ & ~(MF2_DONTDRAW)
	if (overlay.tracer and overlay.tracer.valid)
		overlay.tracer.flags2 = $ & ~(MF2_DONTDRAW)
	end
	
	return false
end
local overlays = {
MT_ELEMENTAL_ORB, MT_FORCE_ORB, MT_ATTRACT_ORB, MT_ARMAGEDDON_ORB, MT_WHIRLWIND_ORB, MT_PITY_ORB, 
MT_FLAMEAURA_ORB, MT_BUBBLEWRAP_ORB, MT_THUNDERCOIN_ORB, 
//MT_OVERLAY	-- needs its own thinker for this
}
for i = 1, #overlays
	addHook("MobjThinker", espio_hideOverlay, overlays[i])
end
addHook("MobjThinker", function(overlay)
	if (overlay.target and overlay.target.valid)
	and (overlay.target.flags2 & MF2_SHIELD)	-- this is a 2nd layer for a shield
		-- we'll handle this inside that thinker instead
		return
	end
	
	-- now just run this like normal
	espio_hideOverlay(overlay)
end, MT_OVERLAY)