-- activate the charge jump!
espio.AbilitySpecial = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (espio.AbilityCheck(p)) then return end
	
	if (mo.espio_chargejump) then return end	-- nvm we're already doing it
	if (p.pflags & PF_THOKKED) then return end	-- thokked state
	if (p.gotcrystal) then return end	-- battlemod support
	if (CBW_Battle and p.exhaustmeter == 0) then return end	-- we're exhausted (battlemod)
	
	p.pflags = ($|PF_THOKKED) & ~(PF_STARTJUMP|PF_SPINNING)
	mo.espio_chargejump = true -- we'll need this
	mo.state = S_ESPIO_CHARGEJUMP	-- set state
	S_StartSound(mo, sfx_s3kab)
end

-- expose these for battle support
espio.ResetChargeJump = function(mo)
	mo.espio_chargejump = nil
	mo.espio_chargejump_timer = nil
	mo.espio_chargejump_halved, mo.espio_chargejump_maxed = nil, nil
	mo.espio_chargejump_frame, mo.espio_chargejump_frametics = nil, nil	-- god...
	mo.espio_chargejump_visualtics = nil
end
espio.StopChargeSounds = function(mo)
	local sounds = {sfx_s3kab, sfx_s3kab1, sfx_s3kab2}
	for i = 1, #sounds
		local sound = sounds[i]
		S_StopSoundByID(mo, sound)
	end
end

local function espio_chargeJumpVisuals(p)
	local mo = p.mo
	if not (p.pflags & PF_JUMPED)
		mo.espio_chargejump_visualtics = nil
		return
	end
	
	local visualtics = mo.espio_chargejump_visualtics
	if (visualtics)
		if not (visualtics%5)
			local dust = P_SpawnMobjFromMobj(mo, 0, 0, -12*FRACUNIT, MT_SPINDUST)
			
			-- change to bubbles underwater
			if (mo.eflags & MFE_UNDERWATER) 
				dust.state = S_SPINDUST_BUBBLE1
			end
		end
		
		-- lower this
		mo.espio_chargejump_visualtics = $-1
	end
end

espio.ChargeJump = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	if not (espio.AbilityCheck(p)) then espio.ResetChargeJump(mo) return end
	
	-- visuals
	espio_chargeJumpVisuals(p)
	
	if not (mo.espio_chargejump) then return end	-- dont do anything
	if (mo.state != S_ESPIO_CHARGEJUMP)	-- we got knocked out of it
	or (p.gotflag or p.gotcrystal)	-- we got the flag (or the crystal in battlemod)
	or (CBW_Battle and p.exhaustmeter == 0)	-- we're exhausted (battlemod)
		if (mo.state == S_ESPIO_CHARGEJUMP)
			-- doesn't reset on its own
			mo.state = S_PLAY_ROLL
		end
		
		espio.ResetChargeJump(mo)
		espio.StopChargeSounds(mo)
		return
	end
	
	mo.espio_chargejump_timer = $ and $+1 or 1
	local timer = mo.espio_chargejump_timer
	
	-- reduce momentum while charging
	local moms = {"momx", "momy"}
	for i = 1, 2
		local mom = moms[i]
		mo[mom] = FixedMul($, espio.ChargeJumpXYMultiplier)
	end
	
	-- slowfall
	local threshold = espio.ChargeJumpSlowfallThreshold
	
	if (mo.eflags & MFE_VERTICALFLIP)
		if not (mo.momz < -(threshold)) 
			mo.momz = min(-(threshold), FixedMul($, espio.ChargeJumpZMultiplier))
		end
	else 
		if not (mo.momz > threshold) 
			mo.momz = min(threshold, FixedMul($, espio.ChargeJumpZMultiplier))
		end
	end
	
	-- calculations + release
	local chargemax = espio.ChargeJumpMax
	if (P_PlayerFullbright(p)) then chargemax = FixedMul($, espio.ChargeJumpSuperMultiplier) end
	-- ^ increase max charge with super espio
	
	local range = (chargemax - espio.ChargeJumpMin)	-- we'll need the range for calcs
	local charge = min(chargemax, espio.ChargeJumpMin + ((range/espio.ChargeJumpDuration)*timer))
	
	-- sounds to play at half and max charge
	if (charge >= (chargemax - (range/2)) and not mo.espio_chargejump_halved)
	or (charge == chargemax and not mo.espio_chargejump_maxed)
		-- chargejump sounds
		espio.StopChargeSounds(mo)
		local sound = (charge != chargemax) and sfx_s3kab1 or sfx_s3kab2
		S_StartSound(mo, sound)
		
		if (charge != chargemax)
			mo.espio_chargejump_halved = true
		else
			mo.espio_chargejump_maxed = true
			S_StartSound(mo, sfx_cdfm50)
		end
	end
	
	-- frame handler...
	local numframes = 6
	mo.espio_chargejump_frame = $ or 0
	mo.espio_chargejump_frametics = $ or 3
	if (charge >= (chargemax - (range/2)))
		mo.espio_chargejump_frametics = min($, (charge == chargemax) and 1 or 2)
	end
	
	-- reduce frame tics and get the frame we want
	mo.espio_chargejump_frametics = $-1
	if not (mo.espio_chargejump_frametics)
		mo.espio_chargejump_frame = $+1
		if (mo.espio_chargejump_frame > (numframes - 1))
			mo.espio_chargejump_frame = 0
		end
	end
	espio.SetFrame(mo, mo.espio_chargejump_frame, SPR2_ROLL)
	
	if not (p.cmd.buttons & BT_JUMP)
		mo.state = S_PLAY_ROLL
		P_SetObjectMomZ(mo, FixedMul(p.jumpfactor, charge)/((mo.eflags & MFE_UNDERWATER) and 2 or 1))
		S_StartSound(mo, sfx_kc5b)
		espio.StopChargeSounds(mo)
		espio.ResetChargeJump(mo)
		
		-- visuals
		local dusts = 8
		for i = 1, dusts
			local angle = ANG1*(360/dusts*i)
			local dust = P_SpawnMobjFromMobj(mo, 32*cos(angle), 32*sin(angle), -5*FRACUNIT, MT_SPINDUST)
			if (mo.eflags & MFE_UNDERWATER) 
				dust.state = S_WPLD1 
				dust.scale = $/2
			end
			P_InstaThrust(dust, angle, FixedMul(mo.scale, 28*FRACUNIT))
		end
		mo.espio_chargejump_visualtics = espio.ChargeJumpVisualTics
	end
end