-- Everything related to colour shifting

-- command so people can set their own shift colours!
COM_AddCommand(espio.ColourShiftCommand, function(p, ...)
	local shiftColours = {}
	local colours = {...}
	if not (#colours) then p.espio_shiftcolours = espio.ColourShiftDefault return end
	
	for k, v in ipairs(colours)
		local colour = R_GetColorByName(v)
		if not (colour) then continue end
		
		table.insert(shiftColours, colour)
	end
	if not (#shiftColours) then return end
	
	p.espio_shiftcolours = shiftColours
	espio.WriteConfig(p)	-- update config
end)

-- the function that handles it
espio.ColourShift = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if (mo.skin != "espio") then mo.espio_shifttimer = nil return end
	
	-- conditions where we should *not* colour shift
	if ((gametyperules & GTR_TEAMS) and p.ctfteam)	-- no confusion in team gamemodes
	or (mo.frame & FF_TRANSMASK) -- multiple ghosts stacking will make us more opaque, so lets prevent it by removing it outright
		mo.espio_shifttimer = nil
		return
	end
	
	-- conditions when we should start colour shifting
	if ((mo.state == S_PLAY_WAIT) or (mo.state == S_PLAY_NIGHTS_STAND)	-- we're bored or standing in nights
	and (mo.frame > 0))	-- and we're not in the transition frame
	//or (mo.espio_focuscharges == 5)	-- we're fully charged!
	or (mo.espio_chargejump_timer and mo.espio_chargejump_timer > espio.ChargeJumpDuration)	-- we're fully charged!
	or (mo.state == S_PLAY_EDGE)	-- we're on the edge
	or (Tether and mo.state == S_PLAY_TETHER_HOLD)	-- smart tether support!
		if (mo.espio_colourshift_delay)
			mo.espio_colourshift_delay = $-1
		else
			mo.espio_shifttimer = $ or 0
		end
	end
	
	if (mo.espio_shifttimer != nil)	-- if it's not nil, we can start
		-- get the colour table
		p.espio_shiftcolours = $ or espio.ColourShiftDefault
		
		-- handle the timer and get maxindex
		mo.espio_shifttimer = $+1
		local shifttics = espio.ColourShiftTics
		local maxindex = #p.espio_shiftcolours + 1
		
		-- the start colour, what we're fading FROM
		local startindex = (mo.espio_shifttimer / shifttics) % maxindex	-- get startindex
		local start = P_SpawnGhostMobj(mo)
		start.frame = ($ & ~FF_TRANSMASK)
		start.rollangle = mo.rollangle
		start.color = p.espio_shiftcolours[startindex] or mo.color
		start.tics = 1
		
		-- the shift colour, what we're fading TO
		local fade = max(1, min(10, 10 - (((mo.espio_shifttimer % shifttics) * 10) / shifttics)))
		
		local shift	-- define here so it's usable outside the block
		if (fade >= 1 and fade <= 9)
			local shiftindex = ((mo.espio_shifttimer + shifttics) / shifttics) % maxindex	-- get shiftindex
			shift = P_SpawnGhostMobj(mo)
			shift.frame = ($ & ~FF_TRANSMASK) | (fade << FF_TRANSSHIFT)
			shift.rollangle = mo.rollangle
			shift.color = p.espio_shiftcolours[shiftindex] or mo.color
			shift.tics = 1
		end
		
		-- hide them if we arent visible
		if (mo.flags2 & MF2_DONTDRAW)	
			-- we want to hide them instead of preventing them from being spawned to prevent desync
			start.flags2 = $|MF2_DONTDRAW
			if (shift and shift.valid)
				shift.flags2 = $|MF2_DONTDRAW
			end
		end
		
		if (mo.espio_shifttimer == ((shifttics * maxindex) - 1))
			mo.espio_shifttimer = nil
		end
	end
end