-- defining everything for espio

rawset(_G, "espio", {})	-- real

rawset(_G, "CA_ESPIOREAL", 95)

-- debug
espio.Debug = false

-- colour shift constants
espio.ColourShiftCommand = "espioshiftcolors"
espio.ColourShiftTics = 20	-- how long it takes to go from colour to colour
espio.ColourShiftDefault = {SKINCOLOR_SAPPHIRE, SKINCOLOR_APPLE, SKINCOLOR_YELLOW, SKINCOLOR_SALMON}

-- focus jump constants
espio.FocusJumpBase = 8*FRACUNIT	-- jump height with no charge
espio.FocusJumpAdd = 9*FRACUNIT/4	-- how much height we get per charge level

-- charge jump constants
espio.ChargeJumpXYMultiplier = FRACUNIT*95/100	-- how much we slow down while charging (horizontally)
espio.ChargeJumpZMultiplier = FRACUNIT*9/10	-- ditto, but vertically
espio.ChargeJumpSlowfallThreshold = -FRACUNIT*15/10
espio.ChargeJumpMin = FRACUNIT*9	-- min power
espio.ChargeJumpMax = FRACUNIT*20	-- max power
espio.ChargeJumpSuperMultiplier = FRACUNIT*12/10	-- increased strength for super as a bonus
espio.ChargeJumpDuration = 9/5*TICRATE	-- how long it takes to get from min to max
espio.ChargeJumpVisualTics = 19

-- wall run constants
espio.WallRunAdd = FRACUNIT*2/5
espio.WallRunTics = 5*TICRATE
espio.WallRunSpeedModifier = FRACUNIT/2
espio.WallRunSpeedSlide = FRACUNIT/5
espio.WallRunTiredThreshold = FRACUNIT*3/8
espio.WallRunJumpSpeed = 7*FRACUNIT
espio.WallRunJumpMinHeight = 10*FRACUNIT
espio.WallRunDelay = 16

-- ceiling cling constants
espio.CeilingClingRange = FRACUNIT*20

-- battle mod constants
espio.BattleExhaustTime = 4*TICRATE

-- invisibility constants (battle mod)
espio.InvisModeCommand = "espioinvismode"
espio.InvisWalkMultiplier = FRACUNIT*3/4
espio.InvisWalkThreshold = FRACUNIT*3/5
espio.InvisRingDrainTicBounds = {TICRATE*2/3, TICRATE}	-- {fastest drain, slowest drain}
espio.InvisNoRingFade = 5
espio.InvisTransparencyAdjust = 6
espio.InvisCooldown = TICRATE/2

-- leaf swirl constants (battle mod)
espio.LeafSwirlMultiplier = FRACUNIT*78/100
espio.LeafSwirlFrameTics = {4, 2, 2}	-- how many tics each frame before the final tic persists
espio.LeafSwirlDuration = TICRATE*6/5	-- how long we're in the animation for
espio.LeafSwirlTornadoSpawnOffset = -56*FRACUNIT
espio.LeafSwirlTornadoTics = 3*TICRATE	-- how long the tornado persists
espio.LeafSwirlDamageRange = 64*FRACUNIT
espio.LeafSwirlLaunchHeight = 16*FRACUNIT
espio.LeafSwirlCooldown = TICRATE*5/2
espio.LeafSwirlLeafOffset = 24	-- we dont want this in fracunits

-- charge thrust constants (battlemod)
espio.ChargeThrustMin = FRACUNIT*28	-- min power
espio.ChargeThrustMax = FRACUNIT*49	-- max power
espio.ChargeThrustMinCooldown = TICRATE
espio.ChargeThrustMaxCooldown = TICRATE*7/4

-- I/O constants
espio.ConfigName = "client/espioconfig.cfg"

-- make sure we have the right ability
espio.AbilityCheck = function(p)
	local mo = p.mo
	
	-- man
	local abilitycheck = p.charability != CA_ESPIOREAL
	if (srb2p) and (mo and mo.valid) then abilitycheck = (mo.skin != "espio") end
	
	if (abilitycheck and not p.espio_wallrun)	-- we use CA_GLIDEANDCLIMB hacks for wall runs, so we need a buffer
		return false
	end
	
	-- we have espio's ability if we're here
	return true
end

-- helper function for setting frames (accounts for frameflags and other extra stuff)
espio.SetFrame = function(mo, frame, sprite, noflags)
	local frameflags = not (noflags) and (mo.frame & ~FF_FRAMEMASK) or 0
	-- ^ if noflags is checked, dont keep frameflags when setting the frame
	
	if (sprite)
		local spr = (mo.player) and "sprite2" or "sprite"	-- players use sprite2
		mo[spr] = sprite
	end
	mo.frame = frame|frameflags
end

-- espio has claimed these sounds
-- all your sounds are belong to us
sfxinfo[sfx_cdfm50].caption = "Charged"
sfxinfo[sfx_kc5b].caption = "Launch"
sfxinfo[sfx_kc45].caption = "Cling"
sfxinfo[sfx_s251].caption = "Dust"

-- espio roll state
freeslot("SPR2_EROL", "S_ESPIO_ROLL")
spr2defaults[SPR2_EROL] = SPR2_ROLL
states[S_ESPIO_ROLL] = {SPR_PLAY, A|SPR2_EROL, 1, nil, 0, 0, S_ESPIO_ROLL}

-- espio chargejump state
freeslot("S_ESPIO_CHARGEJUMP")
states[S_ESPIO_CHARGEJUMP] = {SPR_PLAY, A|SPR2_ROLL, 4, nil, 0, 0, S_ESPIO_CHARGEJUMP}

-- battle mod support
freeslot("MT_ESPIO_SOUNDMOBJ")
mobjinfo[MT_ESPIO_SOUNDMOBJ] = {
	spawnstate = S_INVISIBLE,
	deathstate = S_NULL,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT,
}
addHook("MobjThinker", function(mo)
	if not (mo.soundplaying) then P_RemoveMobj(mo) return end
	if not (S_SoundPlaying(mo, mo.soundplaying)) then P_RemoveMobj(mo) return end
end, MT_ESPIO_SOUNDMOBJ)
espio.SpawnSoundMobj = function(mo, sound)
	local soundmo = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_ESPIO_SOUNDMOBJ)
	soundmo.soundplaying = sound
	S_StartSound(soundmo, sound)
	return soundmo
end

freeslot("S_ESPIO_LEAFSWIRL")
states[S_ESPIO_LEAFSWIRL] = {SPR_PLAY, A|SPR2_TRNS, 3, nil, 0, 0, S_ESPIO_LEAFSWIRL}

freeslot("S_ESPIO_LEAFSWIRL_TORNADO", "MT_ESPIO_LEAFSWIRL_TORNADO")
states[S_ESPIO_LEAFSWIRL_TORNADO] = {SPR_NULL, 0, espio.LeafSwirlTornadoTics + 2, nil, 0, 0, S_NULL}
mobjinfo[MT_ESPIO_LEAFSWIRL_TORNADO] = {
	spawnstate = S_ESPIO_LEAFSWIRL_TORNADO,
	deathstate = S_NULL,
	deathsound = sfx_s3k4b,
	radius = espio.LeafSwirlDamageRange,
	height = espio.LeafSwirlLaunchHeight,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT,
}

-- instead of just using the ACZ dust layers, we'll have our own since for some reason they have 256*FRACUNIT HEIGHT... WHAT?!
for i = 1, 5 do freeslot("S_ESPIO_LEAFSWIRL_DUSTLAYER"..i) end
freeslot("MT_ESPIO_LEAFSWIRL_DUSTLAYER", "SPR_ESDL")
states[S_ESPIO_LEAFSWIRL_DUSTLAYER1] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS70, 2, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER2}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER2] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS70, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER3}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER3] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS80, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER4}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER4] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS80, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER5}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER5] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS90, 4, nil, 0, 0, S_NULL}
mobjinfo[MT_ESPIO_LEAFSWIRL_DUSTLAYER] = {
	spawnstate = S_ESPIO_LEAFSWIRL_DUSTLAYER1,
	deathstate = S_NULL,
	height = 22*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOBLOCKMAP|MF_NOCLIP|MF_SCENERY,
}
for i = 1, 5 do freeslot("S_ESPIO_LEAFSWIRL_LEAF"..i) end
states[S_ESPIO_LEAFSWIRL_LEAF1] = {SPR_ESDL, B|FF_PAPERSPRITE, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF2}
states[S_ESPIO_LEAFSWIRL_LEAF2] = {SPR_ESDL, B|FF_PAPERSPRITE, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF3}
states[S_ESPIO_LEAFSWIRL_LEAF3] = {SPR_ESDL, B|FF_PAPERSPRITE|FF_TRANS30, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF4}
states[S_ESPIO_LEAFSWIRL_LEAF4] = {SPR_ESDL, B|FF_PAPERSPRITE|FF_TRANS60, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF5}
states[S_ESPIO_LEAFSWIRL_LEAF5] = {SPR_ESDL, B|FF_PAPERSPRITE|FF_TRANS90, 4, nil, 0, 0, S_NULL}

addHook("MobjThinker", function(mo)
	local flippedtics = espio.LeafSwirlTornadoTics - mo.tics
	-- ^ convenience vars
	
	-- conditions to destroy the leaf swirl tornado!
	if (mo.tics > 3)	
		-- parry hack
		local target_parrycheck
		
		if ((mo.target and mo.target.valid) and mo.target.health)
		and (mo.target.player.tumble and mo.target.player.tumble > 0)
		and (mo.target.espio_leafswirl_parryhack)
			target_parrycheck = true
		end
		
		-- run conditions
		if not (mo.target and mo.target.valid)	-- target is gone??
		or not (mo.target.health)	-- target is death
		or (target_parrycheck)	-- target got parried
			mo.tics = 3
		end
	end
	
	-- hitbox table {bottom, top}
	local tornadohitboxes
	
	-- rotate spawned layers with a chain system
	local offset = espio.LeafSwirlLeafOffset
	local layer = mo.tracer
	while (layer and layer.valid)
		local leaf = layer.espio_leaf	
		if (layer.espio_leaf)
			-- leaves spawn further out
			layer.espio_leaf = $+2
			leaf = layer.espio_leaf
		end
		
		-- rotate around the center (the tornado)
		local xy_offset = (leaf) and (offset + leaf) or 0
		
		layer.angle = $ + (ANG1*(leaf and 9 or 14))	-- rotate it a tiny bit
		local angle = layer.angle - ANGLE_90	-- the layers are all offset forward by ANGLE_90
		local x = mo.x + (((mo.radius/mo.scale) + xy_offset) * FixedMul(mo.scale, cos(angle)))
		local y = mo.y + (((mo.radius/mo.scale) + xy_offset) * FixedMul(mo.scale, sin(angle)))
		P_TeleportMove(layer, x, y, layer.z)
		if (leaf) then layer.rollangle = $ + ANG1 end
		
		//P_TeleportMove(layer, layer.x + (15 * FixedMul(mo.scale, cos(layer.angle))), layer.y + (15 * FixedMul(mo.scale, sin(layer.angle))), layer.z)
		//layer.scale = mo.scale
		//layer.angle = $ + (ANG1*14)
		layer.momx = mo.momx
		layer.momy = mo.momy
		
		-- make them stop at the ceiling
		if not (leaf)	-- unless it's a leaf
			if (layer.momz)
				local checkz = layer.z + layer.height
				local checkheight = P_CeilingzAtPos(mo.x, mo.y, mo.z, 1)
				if (mo.eflags & MFE_VERTICALFLIP)
					checkz = layer.z
					checkheight = P_FloorzAtPos(mo.x, mo.y, mo.z, 1)
				end
				//print("CheckZ: "..checkz)
				//print("CheckHeight: "..checkheight)
				
				if ((mo.eflags & MFE_VERTICALFLIP) and checkz <= checkheight)
				or (not (mo.eflags & MFE_VERTICALFLIP) and checkz >= checkheight)
					layer.momz = 0
					
					local snapz = (mo.eflags & MFE_VERTICALFLIP) and checkheight or (checkheight - layer.height)
					P_TeleportMove(layer, layer.x, layer.y, snapz)
					//print("stop")
				end
			end
			
			-- set hitbox
			local layertop = layer.z + layer.height
			local layerbottom = layer.z
			
			if (tornadohitboxes == nil)	-- we dont have hitboxes set
				tornadohitboxes = {layerbottom, layertop}
			else	-- we already got one, check heights to reset it
				-- bottom hitbox
				if (tornadohitboxes[1] > layerbottom)
					tornadohitboxes[1] = layerbottom
				end
				
				-- top hitbox
				if (tornadohitboxes[2] < layertop)
					tornadohitboxes[2] = layertop
				end
			end
		end
			
		-- go to the next layer in the chain (if there is one)
		layer = layer.tracer
	end
	
	-- spawn layers
	if not (mo.extravalue1%6)
	and (mo.tics > 2)
		for i = 1, 4
			local angle = mo.angle + ANGLE_90*i
			local x = (mo.radius/mo.scale) * cos(angle)
			local y = (mo.radius/mo.scale) * sin(angle)
		
			local layer = P_SpawnMobjFromMobj(mo, x, y, 0, MT_ESPIO_LEAFSWIRL_DUSTLAYER)
			layer.angle = angle + ANGLE_90
			layer.color = (G_GametypeHasTeams()) and mo.target.color or SKINCOLOR_WHITE
			layer.flags = $|(MF_NOCLIPHEIGHT)
			P_SetObjectMomZ(layer, 12*FRACUNIT)
			
			-- set hitboxes here too
			local layertop = layer.z + layer.height
			local layerbottom = layer.z
			
			if (tornadohitboxes == nil)	-- we dont have hitboxes set
				tornadohitboxes = {layerbottom, layertop}
			else	-- we already got one, check heights to reset it
				-- bottom hitbox
				if (tornadohitboxes[1] > layerbottom)
					tornadohitboxes[1] = layerbottom
				end
				
				-- top hitbox
				if (tornadohitboxes[2] < layertop)
					tornadohitboxes[2] = layertop
				end
			end
			
			-- chain?
			layer.tracer = mo.tracer
			mo.tracer = layer
		end
		
		-- spawn leaves
		local leaves = 6
		for i = 1, leaves
			local angle = mo.angle + ((ANG1*(360/leaves))*i)
			local x = ((mo.radius/mo.scale) + offset) * cos(angle)
			local y = ((mo.radius/mo.scale) + offset) * sin(angle)
			
			local layer = P_SpawnMobjFromMobj(mo, x, y, FRACUNIT*24, MT_ESPIO_LEAFSWIRL_DUSTLAYER)
			layer.state = S_ESPIO_LEAFSWIRL_LEAF1
			layer.angle = angle + ANGLE_90
			layer.color = (G_GametypeHasTeams()) and mo.target.color or SKINCOLOR_EMERALD
			layer.flags = $|(MF_NOCLIPHEIGHT)
			P_SetObjectMomZ(layer, 7*FRACUNIT)
			
			-- chain?
			layer.espio_leaf = 1	-- just so we can distinguish these to make them spaced more outwards
			layer.tracer = mo.tracer
			mo.tracer = layer
		end
	end
	mo.extravalue1 = $+1
	
	-- damage stuff around us
	local function espio_tornadoBlockmap(ref, found)
		if not ((found and found.valid) and found.health) then return end
		if (ref == found) then return end	-- brew
		
		if (found == ref.target) then return end	-- cant damage ourself!
		if not (P_CheckSight(ref, found)) then return end	-- self explanatory
		if (found.player)	-- we gotta do our own checks here
			local B = CBW_Battle
			
			if not (B.PlayerCanBeDamaged(found.player)) then return end	-- cant be damaged
			if (B.MyTeam(found.player, ref.target.player)) then return end	-- same team
		else
			if not (found.flags & (MF_SHOOTABLE|MF_ENEMY|MF_MONITOR)) then return end	-- damage enemies and monitors!
		end
		if not ((tornadohitboxes[2] >= found.z) and (tornadohitboxes[1] <= found.z + found.height)) then return end	-- z checks
		
		-- ACTUAL distance check, so it's ciruclar and not square
		local dist = mo.radius + FixedMul(mo.scale, 16*FRACUNIT)
		if not (R_PointToDist2(found.x, found.y, ref.x, ref.y) <= dist) then return end
		
		-- if they're in the leaf swirl actionstate, make them the inflictor so they can be parried out!
		local inflictor = (ref.target.player.actionstate == 2) and ref.target or ref
		P_DamageMobj(found, inflictor, ref.target)
		
		if ((found and found.valid) and found.health)	-- doublecheck this, we only want to launch them if they're still alive!
		and not (found.player and not P_PlayerInPain(found.player))	-- assume a parry happened!
			-- make sure they're launched AWAY from the tornado
			found.angle = R_PointToAngle2(found.x, found.y, ref.x, ref.y)
			P_InstaThrust(found, found.angle, -FixedHypot(found.momx, found.momy))
			
			-- launch them!
			if not (found.flags & MF_BOSS)
				P_SetObjectMomZ(found, ref.height)
			end
		end
	end
	
	if (tornadohitboxes != nil)
		searchBlockmap("objects", espio_tornadoBlockmap, mo, mo.x - mo.radius, mo.x + mo.radius, mo.y - mo.radius, mo.y + mo.radius)
		
		-- hitbox check
		if (espio.Debug)
			if not (mo.debugthoks)
				mo.debugthoks = {}
				
				for i = 1, 2
					local thok = P_SpawnMobj(mo.x, mo.y, tornadohitboxes[i], MT_THOK)
					thok.frame = ($|FF_FULLBRIGHT) & ~FF_TRANSMASK	-- give fullbright, remove transparency
					thok.color = (i == 2) and SKINCOLOR_WHITE or SKINCOLOR_BLACK
					thok.flags = $ & ~(MF_SHOOTABLE|MF_ENEMY|MF_MONITOR)	-- we dont want our debug thoks to be killed
					thok.flags = $|(MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT)
					thok.tics = -1
					
					thok.debugthoknum = i
					table.insert(mo.debugthoks, thok)
				end
			else
				for k, v in pairs(mo.debugthoks)
					P_TeleportMove(v, mo.x, mo.y, tornadohitboxes[v.debugthoknum])
				end
			end
		end
	else
		if (espio.Debug)
			if (mo.debugthoks)
				for k, v in pairs(mo.debugthoks)
					P_RemoveMobj(v)
				end
				mo.debugthoks = nil
			end
		end
	end
	
	-- play sounds
	if not (S_SoundPlaying(mo, sfx_s3kcel)) then S_StartSound(mo, sfx_s3kcel) end	-- active sound
	if (mo.tics == 3)	-- death sound
		-- doing this so it doesnt get cut off
		local soundmo = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		soundmo.state = S_INVISIBLE
		soundmo.fuse = 3*TICRATE
		S_StartSound(soundmo, mo.info.deathsound) 
	end
	
	-- dont despawn if we still have dustlayers!
	if (mo.tics <= 2)	-- cant check 1 because it despawns at 0
	and (mo.tracer and mo.tracer.valid)	-- mo.tracer should always be the last spawn layer
		mo.tics = 2
	end
end, MT_ESPIO_LEAFSWIRL_TORNADO)

-- Xmomentum support
freeslot("SPR2_SHIT")
spr2defaults[SPR2_SHIT] = SPR2_PAIN

-- smart tether support
freeslot("SPR2_HOLD")
spr2defaults[SPR2_HOLD] = SPR2_EDGE

-- srb2p support
freeslot("SPR2_PSMN")	-- have these in SPR2, actually
spr2defaults[SPR2_PSMN] = SPR2_STND

-- srb2p (2) support
freeslot("SPR2_PRTY")