-- handles drilling, visuals AND certain functionality
local function espio_spinTrail(p)
	local mo = p.mo
	
	-- use a normal spin trail in this scenario
	if (mo.state == S_ESPIO_ROLL)
		p.spinitem = MT_THOK
		return
	end
	p.spinitem = MT_NULL	-- reset this just incase
	
	-- dont spawn a trail under these conditions
	if (p.powers[pw_carry] == CR_ZOOMTUBE)	-- in a zoomtube
	or (p.powers[pw_carry] == CR_MACESPIN)	-- on the CEZ maces
		return
	end
	
	if ((p.pflags & PF_SPINNING) and (P_AproxDistance(p.speed, mo.momz) > FixedMul(15<<FRACBITS, mo.scale)) and not (p.pflags & PF_JUMPED))
	and (P_IsObjectOnGround(mo) and (leveltime%2) and not (mo.eflags & MFE_GOOWATER))
		for i = 1, 2
			local dust = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SPINDUST)
			
			-- set state under the right conditions
			if (mo.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER)) 
				dust.state = S_SPINDUST_BUBBLE1
			elseif (p.powers[pw_shield] == SH_ELEMENTAL)
				dust.state = S_SPINDUST_FIRE1
			end
			
			-- set target and scale
			dust.target = mo
			dust.scale = mo.scale*2/3
			
			-- momentum
			local momz = 2*FRACUNIT + (P_RandomFixed()<<2)
			local angle = R_PointToAngle2(0, 0, mo.momx, mo.momy) + (ANG1*P_RandomRange(-50, 50))
			local speed = FRACUNIT + (P_RandomFixed()<<3)
			
			P_SetObjectMomZ(dust, momz)
			P_InstaThrust(dust, angle, -(speed))
			P_TryMove(dust, dust.x + dust.momx, dust.y + dust.momy, true)
		end
	end
end

local function espio_resetBuster(mo, dashspeed)
	if (dashspeed)
		mo.espio_fakedashspeed = nil
	end
	
	if (mo.espio_debrisobject)
		mo.espio_debrisobject = nil
	end
end

-- handles the spindash drilling into bustable floors
local function espio_floorBuster(p)
	local mo = p.mo
	if not (p.pflags & PF_STARTDASH) then espio_resetBuster(mo, true) return end
	
	-- get the rover we'll be checking
	local rover = mo.floorrover
	if (mo.eflags & MFE_VERTICALFLIP) then rover = mo.ceilingrover end
	
	if not (rover and rover.valid) then espio_resetBuster(mo) return end
	if not (rover.flags & FF_EXISTS) then espio_resetBuster(mo) return end
	if not ((rover.flags & FF_STRONGBUST) or (rover.flags & FF_BUSTUP)) then espio_resetBuster(mo) return end
	-- ^ after all these checks we should be good
		
	-- init var
	mo.espio_fakedashspeed = $ or 0
		
	local soundcalculation = (6*(mo.espio_fakedashspeed - p.mindash))/(p.maxdash - p.mindash)
	mo.espio_fakedashspeed = max(p.mindash, min(p.maxdash, $+FRACUNIT))
			
	-- redefine calculation then play spindash sound
	local chargecalculation = (6*(mo.espio_fakedashspeed - p.mindash))/(p.maxdash - p.mindash)
	if (soundcalculation != chargecalculation)
		-- try to get the object?
		if not (mo.espio_debrisobject)	-- this is actually a really laggy method so dont do this too much
		end
		
		-- spawn dirt from player
		for i = 1, 5
			local rock = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_PLAYER)
			rock.scale = mo.scale/3
			rock.sprite = SPR_ROIA
			rock.flags = MF_SCENERY
			rock.fuse = TICRATE*3/4
			P_InstaThrust(rock, ANG1*P_RandomRange(1, 360), P_RandomRange(2, 5)*rock.scale)
			P_SetObjectMomZ(rock, P_RandomRange(6, 12)*FRACUNIT)
		end
		
		S_StartSound(mo, sfx_crumbl)	-- time to revv the engine!
	end
	
	-- use normal roll frames
	if (mo.skin == "espio")
	and (mo.sprite2 == SPR2_SPIN)
		mo.sprite2 = SPR2_ROLL
		
		local step = (p.maxdash - p.mindash)/4
		local speed = (p.dashspeed - p.mindash)
		if (speed > 3*step)
			espio.SetFrame(mo, leveltime%6)
		elseif (speed > step)
			espio.SetFrame(mo, (leveltime%12)/2)
		else
			espio.SetFrame(mo, (leveltime%18)/3)
		end
	end
	
	-- break the bustable floor on max charge!
	if (mo.espio_fakedashspeed != p.maxdash) then return end
	
	-- bust the floor
	EV_CrumbleChain(rover)
	if (rover.master.flags & ML_EFFECT5)
		P_LinedefExecute(P_AproxDistance(rover.master.dx, rover.master.dy)>>FRACBITS, mo, mo.subsector.sector)
	end
	
	-- set player state
	p.pflags = $|(PF_JUMPED|PF_THOKKED|PF_SPINNING)
	mo.state = S_PLAY_ROLL
	P_SetObjectMomZ(mo, 5*FRACUNIT/((mo.eflags & MFE_UNDERWATER) and 2 or 1))
	S_StartSound(mo, sfx_crumbl)	-- time to revv the engine!
end

-- functions for removing shieldability in specific scenarios
espio.shieldAbilityFuncs = {
	-- have this always up during attraction shield's shield ability
	[SH_ATTRACT] = {
		removefunc = function(p)
			local mo = p.mo
			
			if not (p.pflags & PF_SHIELDABILITY) 
				mo.espio_rolling_shieldability = nil
			end
		end,
	},
	
	-- ditto, for the flame shield
	[SH_FLAMEAURA] = {
		removefunc = function(p)
			local mo = p.mo
			
			if not (p.pflags & PF_SHIELDABILITY) 
				mo.espio_rolling_shieldability = nil
			end
		end,
	},
	
	[SH_BUBBLEWRAP] = {
		removefunc = function(p)
			local mo = p.mo
			
			if not (p.pflags & PF_SHIELDABILITY)
			and not (p.secondjump)	-- this SHOULD be for the bubble shield's bounce?
				mo.espio_rolling_shieldability = nil
			end
		end,
	},
	
	[SH_THUNDERCOIN] = {
		func = function(p)
			local mo = p.mo
			
			-- just to delay it a bit
			mo.espio_rolling_delayshieldability = true
		end,
		
		removefunc = function(p)
			local mo = p.mo
			
			if (mo.espio_rolling_delayshieldability)
				mo.espio_rolling_delayshieldability = nil
				return
			end
			
			-- NOW we reset it
			mo.espio_rolling_shieldability = nil
		end,
	},
}

espio.ShieldSpecial = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	if (mo.skin != "espio") then return end
	
	local shield = (p.powers[pw_shield] & SH_NOSTACK)
	if (shield and (shield & SH_FORCE)) then shield = SH_FORCE end
	if not (shield) then return end	-- ?????????????
	
	-- these dont have shield abilities, or we dont want to curl
	local invalidShields = {SH_PITY, SH_PINK, SH_WHIRLWIND}
	for k, v in pairs(invalidShields)
		if (v == shield) then return end
	end
	
	mo.state = S_ESPIO_ROLL
	mo.espio_rolling_shieldability = shield
	if (espio.shieldAbilityFuncs[shield])
	and (espio.shieldAbilityFuncs[shield].func)
		espio.shieldAbilityFuncs[shield].func(p)
	end
end

local function espio_handleShieldAbility(p)
	local mo = p.mo
	local shieldability = mo.espio_rolling_shieldability
	if not (shieldability) then return end
	
	if (espio.shieldAbilityFuncs[shieldability])
	and (espio.shieldAbilityFuncs[shieldability].removefunc)
		espio.shieldAbilityFuncs[shieldability].removefunc(p)
	else
		if (mo.state == S_ESPIO_ROLL) then return end
		mo.espio_rolling_shieldability = nil
	end
end

-- with espio's taller spinheight he has a tendency to no interact with things correctly, time to circumvent that
local function espio_startRolling(p)
	local mo = p.mo
	
	if (p.powers[pw_carry] == CR_ZOOMTUBE)	-- in a zoomtube
	or (p.powers[pw_carry] == CR_MACESPIN)	-- on the CEZ maces
	or (mo.espio_rolling_shieldability and (p.pflags & PF_JUMPED))
		p.spinheight = 32*FRACUNIT
		
		if (p.powers[pw_carry] == CR_ZOOMTUBE)
			if (mo.sprite2 == SPR2_EROL) then return end
			
			mo.sprite2 = SPR2_EROL
			
			-- set frame manually (GOD)
			local speed = FixedDiv(p.speed, FixedMul(mo.scale, mo.movefactor))	-- for some reason zoomtubes animate slower if they're slow as well
			espio.SetFrame(mo, (speed > 16<<FRACBITS) and (leveltime%4) or (leveltime%8)/2)
			-- we'll use this to set our state to S_ESPIO_ROLL afterwards
			mo.espio_rolling_zoomtube = true
		else
			if (mo.state == S_ESPIO_ROLL) then return end
			
			mo.state = S_ESPIO_ROLL
		end
	else
		-- end my life
		if (mo.espio_rolling_zoomtube)
			mo.state = S_ESPIO_ROLL
			mo.espio_rolling_zoomtube = nil
		end
		
		if (mo.state == S_ESPIO_ROLL) then return end	-- we're still rolling
		if (mo.sprite2 == SPR2_EROL) then return end	-- we're probably still rolling?
		
		p.spinheight = skins[mo.skin].spinheight	-- reset spinheight
	end
end

local function espio_rollAnimating(p)
	local mo = p.mo
	if (mo.state != S_ESPIO_ROLL) then return end
	
	-- recreate how S_PLAY_ROLL animates
	local speed = FixedDiv(p.speed, FixedMul(mo.scale, mo.movefactor))	-- for some reason zoomtubes animate slower if they're slow as well
	espio.SetFrame(mo, (speed > 16<<FRACBITS) and (leveltime%4) or (leveltime%8)/2)
end

espio.DrillSpin = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	if (mo.skin != "espio") then return end
	
	espio_spinTrail(p)
	espio_floorBuster(p)
	espio_handleShieldAbility(p)
	espio_startRolling(p)
	espio_rollAnimating(p)
end