-- Everything related to Espio's focus jump.
-- Purely here for archival purposes lol
/*
-- handles the orbs that surrounds us
local function espio_removeOrbs(p)
	p.espio_orbs = $ or {}
	if not (#p.espio_orbs) then return end
	
	local i = #p.espio_orbs
	while i
		P_RemoveMobj(p.espio_orbs[i])
		table.remove(p.espio_orbs, i)
		i = $-1
	end
end
local function espio_handleOrbs(p)
	local mo = p.mo
	p.espio_orbs = $ or {}
	if not (#p.espio_orbs) then return end
	
	-- we shouldnt have orbs!
	if not ((mo and mo.valid) and mo.health)
	or not (espio.AbilityCheck(p))
		espio_removeOrbs(p)
		return
	end
	
	-- we have more orbs than we should?
	mo.espio_focuscharges = $ or 0
	while (#p.espio_orbs > mo.espio_focuscharges)
		local orb = p.espio_orbs[#p.espio_orbs]
		if (orb and orb.valid)
			P_RemoveMobj(p.espio_orbs[#p.espio_orbs])
		end
		
		table.remove(p.espio_orbs, #p.espio_orbs)
	end
	if not (#p.espio_orbs) then return end	-- gotta check this again
	
	local offset_values = {{40, 9}, {36, 10}, {30, 12}, {24, 15}, {20, 18}}
	local angleoffset = (ANG1*(leveltime%(offset_values[mo.espio_focuscharges][1]))*(offset_values[mo.espio_focuscharges][2]))
	for k, v in pairs(p.espio_orbs)
		if not (v and v.valid)
			table.remove(p.espio_orbs, k)
			continue
		end
		
		local orbtimer = leveltime + v.espio_leveltime
		local angle = v.espio_orbangle + angleoffset
		local xpos = mo.x + FixedMul(mo.scale, ((mo.radius*7/4)>>FRACBITS)*cos(angle))
		local ypos = mo.y + FixedMul(mo.scale, ((mo.radius*7/4)>>FRACBITS)*sin(angle))
		
		-- teleport the orb
		v.color = p.skincolor
		P_TeleportMove(v, xpos, ypos, mo.z + FixedMul(mo.scale, v.espio_orbheight))
		
		-- spawn ghost on the orb
		if (orbtimer%2)
			local g = P_SpawnGhostMobj(v)
			g.destscale = 1
			g.tics = 5
		end
		
		-- second ghost, this time to make it flash
		
		local fadeValues = {0, 8, 6, 5, 6, 8}
		local fade = fadeValues[((orbtimer%24)/4)+1]
		
		if (fade >= 1 and fade <= 10)
			local flash = P_SpawnGhostMobj(v)
			flash.frame = ($ & ~FF_TRANSMASK) | (fade << FF_TRANSSHIFT)
			flash.color = SKINCOLOR_WHITE
			flash.tics = 1
		end
	end
end

-- adds a charge
local function espio_addCharge(mo, charges)
	local alreadymaxed = (mo.espio_focuscharges == 5)
	
	mo.espio_focuscharges = $ and min(5, $+charges) or charges
	mo.espio_focusjumped = 1	-- reset this
	S_StartSound(mo, (mo.espio_focuscharges == 5) and sfx_cdfm50 or sfx_s3k42)
	
	-- add visual charge orbs
	if (alreadymaxed) then return end
	mo.player.espio_orbs = $ or {}
	for i = 1, charges
		local angle = ANG1*P_RandomRange(1, 360)
		local height = P_RandomRange(8, (mo.height/mo.scale))*FRACUNIT
		local xoff = FixedMul(mo.scale, ((mo.radius*7/4)>>FRACBITS)*cos(angle))
		local yoff = FixedMul(mo.scale, ((mo.radius*7/4)>>FRACBITS)*sin(angle))
		
		local orb = P_SpawnMobjFromMobj(mo, xoff, yoff, FixedMul(mo.scale, height), MT_THOK)
		orb.frame = ($ & ~FF_TRANSMASK)|FF_FULLBRIGHT|FF_TRANS40
		orb.scale = FixedMul(mo.scale, FRACUNIT/18)
		orb.destscale = FixedMul(mo.scale, FRACUNIT/4)
		orb.scalespeed = FRACUNIT/120
		orb.color = mo.player.skincolor
		orb.tics = -1
			
		-- control vars for the table
		orb.espio_orbangle = angle
		orb.espio_orbheight = height
		orb.espio_leveltime = leveltime
		table.insert(mo.player.espio_orbs, orb)
	end
end

-- gain charges when breaking monitors or destroying enemies!
-- first check: enemies
addHook("MobjDamage", function(mo, inflictor, source)
	if not (source and source.valid) then return end
	if (source.type != MT_PLAYER) then return end
	local p = source.player
	
	-- check if it's an enemy
	if not (mo.flags & (MF_ENEMY|MF_BOSS)) then return end
	
	-- if they're espio, add a charge!
	if not (espio.AbilityCheck(p)) then return end
	espio_addCharge(source, 1)
end)
-- second check: monitors
addHook("MobjDeath", function(mo, inflictor, source)
	if not (source and source.valid) then return end
	if (source.type != MT_PLAYER) then return end
	local p = source.player
	
	-- check if it's a monitor
	if not (mo.flags & MF_MONITOR) then return end
	
	-- if they're espio, add a charge!
	if not (espio.AbilityCheck(p)) then return end
	espio_addCharge(source, 1)
end)

-- activate the focus jump!
espio.AbilitySpecial = function(p)
	-- nil: havent used it at all
	-- 1: still able to use it
	-- 2: used it
	-- 3: used a shield ability; cant use it till you land
	
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (espio.AbilityCheck(p)) then return end
	if (mo.espio_focusjumped and mo.espio_focusjumped > 1) then return end
	
	-- get charge and set momz based on that
	local momz = FixedMul(espio.FocusJumpBase, mo.scale)
	if (mo.espio_focuscharges)
		momz = $ + FixedMul(espio.FocusJumpAdd * mo.espio_focuscharges, mo.scale)
		mo.espio_focuscharges = nil	-- remove charge
	end
	mo.momz = (momz*P_MobjFlip(mo))/((mo.eflags & MFE_UNDERWATER) and 2 or 1)
	S_StartSound(mo, sfx_bnce2)
	
	-- set vars, we cant do this again
	p.pflags = $|PF_THOKKED|PF_STARTJUMP	-- add startjump so we can control our jump!
	mo.espio_focusjumped = 2
	espio_removeOrbs(p)
	
	-- set state, if we're rolling after a mace or zoomtube (somehow???)
	if (mo.state == S_ESPIO_ROLL)
		p.spinheight = skins[mo.skin].spinheight
		mo.state = S_PLAY_ROLL
	end
end

espio.FocusJump = function(p)
	-- handle the orbs that surround us here.
	espio_handleOrbs(p)
	
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	if not (espio.AbilityCheck(p))
		mo.espio_focuscharges = nil
		mo.espio_focusjumped = nil
		return 
	end
	
	-- shield abilities are a cutoff point for focus jumping
	if (p.pflags & PF_THOKKED)
	and not (mo.espio_focusjumped)
		mo.espio_focusjumped = 3
	end
	
	-- reset the focus jump if we've landed (or we're not in thokked state!)
	if (P_IsObjectOnGround(mo))
	or not (p.pflags & PF_THOKKED)
		mo.espio_focusjumped = nil
	end
end*/