-- function to write a config (should be called whenever anything is set)
espio.WriteConfig = function(p)
	-- consoleplayer check, limit netgame shenanigans
	if (p != consoleplayer) then return end
	
	-- load the config we'll be writing to!
	local config = io.openlocal(espio.ConfigName, "w+")
	
	-- write espio's shift colours
	local shiftColours = p.espio_shiftcolours or espio.ColourShiftDefault
	local shiftString = espio.ColourShiftCommand.." "
	for k, v in ipairs(shiftColours)
		shiftString = $ + tostring(v)
		if (k != #shiftColours) then shiftString = $ + " " end	-- this is how we'll split this up
	end
	config:write(shiftString)
	config:write("\n")	-- line break
	
	-- write espio's invis mode
	local invisString = espio.InvisModeCommand.." "
	local invisMode = (p.espio_invismode) and tostring(p.espio_invismode) or "1"
	invisString = $ + invisMode
	config:write(invisString)
	
	-- close the config
	config:close()
end

-- load the config once we are valid
local loadedplayer
espio.LoadConfig = function()
	if (consoleplayer and consoleplayer.valid)
	and not (loadedplayer and loadedplayer.valid)	-- this'll let it reload whenever we join a server
		-- load the config we'll be reading from
		local config = io.openlocal(espio.ConfigName, "r")
		if not (config)	-- no config? create one real quick
			espio.WriteConfig(consoleplayer)
			config = io.openlocal(espio.ConfigName, "r")	-- now this shouldn't fail
		end
		
		-- load each line of the config (technically you could stuff other commands in here!)
		for line in config:lines()
			COM_BufInsertText(consoleplayer, line)
		end
		
		-- close the config, we're done
		config:close()
		
		-- save the consoleplayer so this isnt run constantly
		loadedplayer = consoleplayer
	end
end